
#define WINDOWS_H

/****************************************************************

    Copyright (c) 1996-1997 Fred Pesch

    This program is free software; you can redistribute it 
    and/or modify it under the terms of the GNU General 
    Public License as published by the Free Software 
    Foundation (version 2 of the License).

    This program is distributed in the hope that it will 
    be useful, but WITHOUT ANY WARRANTY; without even the 
    implied warranty of MERCHANTABILITY or FITNESS FOR A 
    PARTICULAR PURPOSE.  See the GNU General Public License 
    for more details.

    For a copy of the GNU General Public License, write to the 
    Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
    MA 02139, USA.

****************************************************************/

#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/CascadeB.h>
#include <Xm/PushB.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#include "interface.h"
#include "dirtree.h"
#include "config.h"

Atom _FOF_SPECIFIC_PROPERTY;
Atom _EXODE_CLIENT_PROPERTY_TYPE;
Atom _EXODE_ROOT_PROPERTY_TYPE;
Atom _EXODE_WINDOWSLIST_PROPERTY;

int NbButtons=0;
int MaxButtons=0;
Widget *Buttons=NULL;
Window *WindowIds=NULL;
unsigned long *Pids=NULL;

int (*OldErrorHandler)(Display *, XErrorEvent *);

/* tile, stack & cascade parameters */

unsigned int RWin_LeftOffset=80;
unsigned int RWin_TopOffset=50;
unsigned int RWin_RightOffset=80;
unsigned int RWin_BottomOffset=50;
unsigned int RWin_XInterOffset=10;
unsigned int RWin_YInterOffset=40;
unsigned int RWin_CascadeXOffset=30;
unsigned int RWin_CascadeYOffset=40;

void AddWindowToRootProperty(void)
{
  int result;
  int format;
  Atom ptype;
  unsigned long nb;
  unsigned long left;
  Window* value;
  Window addvalue;
  int i;

  _EXODE_ROOT_PROPERTY_TYPE =  XInternAtom(XtDisplay(W_App), "_EXODE_ROOT_PROPERTY_TYPE", False);

  _EXODE_WINDOWSLIST_PROPERTY=  XInternAtom(XtDisplay(W_App), "_EXODE_WINDOWSLIST_PROPERTY", False);

  /* modify the root windows list property */

  /* get the actual contents of the property */

  result = XGetWindowProperty(XtDisplay(W_App), DefaultRootWindow(XtDisplay(W_App)), _EXODE_WINDOWSLIST_PROPERTY, 0,16384, False, _EXODE_ROOT_PROPERTY_TYPE, &ptype, &format, &nb, &left, (unsigned char **) &value);

  /* if there is something in the property */

  if(result==Success && nb!=0) {
    for(i=0;i<nb;i++)
      if(XtWindow(W_App)==value[i]) {
	/* the window is already in the list */
	XFree ( value);
	return;
      }

    XFree( value);
  }
  
  /* add the window to the list */
  
  addvalue = XtWindow(W_App);

  if(nb==0) 
    XChangeProperty(XtDisplay(W_App), DefaultRootWindow(XtDisplay(W_App)), _EXODE_WINDOWSLIST_PROPERTY, _EXODE_ROOT_PROPERTY_TYPE, 32, PropModeReplace, (unsigned char *) &addvalue, 1);
  else
    XChangeProperty(XtDisplay(W_App), DefaultRootWindow(XtDisplay(W_App)), _EXODE_WINDOWSLIST_PROPERTY, _EXODE_ROOT_PROPERTY_TYPE, 32, PropModeAppend, (unsigned char *) &addvalue, 1);
  
}

void RemoveRootProperty(void)
{
  int result;
  int format;
  Atom ptype;
  unsigned long nb;
  unsigned long left;
  Window* value;

  _EXODE_ROOT_PROPERTY_TYPE =  XInternAtom(XtDisplay(W_App), "_EXODE_ROOT_PROPERTY_TYPE", False);

  _EXODE_WINDOWSLIST_PROPERTY=  XInternAtom(XtDisplay(W_App), "_EXODE_WINDOWSLIST_PROPERTY", False);

  /* modify the root windows list property */

  /* get the actual contents of the property and clear it */

  result = XGetWindowProperty(XtDisplay(W_App), DefaultRootWindow(XtDisplay(W_App)), _EXODE_WINDOWSLIST_PROPERTY, 0,16384, True, _EXODE_ROOT_PROPERTY_TYPE, &ptype, &format, &nb, &left, (unsigned char **) &value);

  /* if there is something in the property */

  if(result!=Success) return;

  XFree( (unsigned char *) value);

}  

void RemoveWindowFromRootProperty(Window wvalue)
{
  int result;
  int format;
  Atom ptype;
  unsigned long nb;
  unsigned long left;
  Window* value;
  Window* newvalue;
  int stop;
  int i,j;

  _EXODE_ROOT_PROPERTY_TYPE =  XInternAtom(XtDisplay(W_App), "_EXODE_ROOT_PROPERTY_TYPE", False);

  _EXODE_WINDOWSLIST_PROPERTY=  XInternAtom(XtDisplay(W_App), "_EXODE_WINDOWSLIST_PROPERTY", False);

  /* modify the root windows list property */

  /* get the actual contents of the property */

  result = XGetWindowProperty(XtDisplay(W_App), DefaultRootWindow(XtDisplay(W_App)), _EXODE_WINDOWSLIST_PROPERTY, 0,16384, False, _EXODE_ROOT_PROPERTY_TYPE, &ptype, &format, &nb, &left, (unsigned char **) &value);

  /* if there is something in the property */

  if(result!=Success) return;

  if(nb==1) {
    XFree ( (unsigned char *) value);
    RemoveRootProperty();
    return;
  }

  stop=1;

  if(nb>0) {
    for(i=0;i<nb;i++)
      if(wvalue==value[i])
	/* the window is in the list */
	stop=0;
  }

  if(stop) {
    XFree( (unsigned char *) value);
    return;
  }

  /* remove the window from the list */

  newvalue = (Window *) XtMalloc(sizeof(Window)*(nb-1));

  i=0;
  j=0;
  while(i<nb) {
    if (wvalue!=value[i])
      newvalue[j++]=value[i];
    i++;
  }

  XChangeProperty(XtDisplay(W_App), DefaultRootWindow(XtDisplay(W_App)), _EXODE_WINDOWSLIST_PROPERTY, _EXODE_ROOT_PROPERTY_TYPE, 32, PropModeReplace, (unsigned char *) newvalue, nb-1);

  XFree( value);
  XtFree( (XtPointer) newvalue);

}  

Window * QueryRootProperty(unsigned long *nb)
{
  int result;
  int format;
  Atom ptype;
  unsigned long left;
  Window* value;

  _EXODE_ROOT_PROPERTY_TYPE =  XInternAtom(XtDisplay(W_App), "_EXODE_ROOT_PROPERTY_TYPE", False);

  _EXODE_WINDOWSLIST_PROPERTY=  XInternAtom(XtDisplay(W_App), "_EXODE_WINDOWSLIST_PROPERTY", False);

  /* get the actual contents of the property */

  result = XGetWindowProperty(XtDisplay(W_App), DefaultRootWindow(XtDisplay(W_App)), _EXODE_WINDOWSLIST_PROPERTY, 0,16384, False, _EXODE_ROOT_PROPERTY_TYPE, &ptype, &format, nb, &left, (unsigned char **) &value);

  /* if there is something in the property */

  if(result!=Success || *nb==0) return NULL;

  return value;

}  

void ChangeWindowProperty(void)
{
  char exode_sign[128];
  char client_sign[128];
  char windowid[128];
  char pid[128];
  char dir[256];
  int nbletters;
  int i,j;

  char text[256+4*128+128];
 
  _FOF_SPECIFIC_PROPERTY = XInternAtom(XtDisplay(W_App), "_FOF_SPECIFIC_PROPERTY", False);

  _EXODE_CLIENT_PROPERTY_TYPE = XInternAtom(XtDisplay(W_App), "_EXODE_CLIENT_PROPERTY_TYPE", False);

  strcpy(exode_sign,"EXODE");
  strcpy(client_sign,"FOF");
  
  sprintf(windowid,"0x%x",(unsigned int) XtWindow(W_App));
  sprintf(pid,"%lu",(unsigned long int) getpid());

  if(strcmp(ActualDir->dir.pathname,"")==0)
    strcpy(dir, "/");
  else
  if(GetVarInt("WindowMenuTextMaxLength", &nbletters)!= (int)NULL) {
    if(nbletters>=strlen(ActualDir->dir.pathname))
      strcpy(dir, ActualDir->dir.pathname);
    else {
      for(i=0;i< (nbletters >> 1)-2 ; i++)
	dir[i]=ActualDir->dir.pathname[i];
      dir[i++]='.';
      dir[i++]='.';
      dir[i++]='.';
      dir[nbletters]=(char)0;
      for(j=nbletters-1;j>=i;j--)
	dir[j]=ActualDir->dir.pathname[strlen(ActualDir->dir.pathname)+j-nbletters+1];
    }
  } else {
    strncpy(dir, ActualDir->dir.pathname, 255);
    dir[255]=(char)0;
  }
  
  strcpy(text, exode_sign);
  strcat(text, "#");
  strcat(text, client_sign);
  strcat(text, "#");
  strcat(text, windowid);
  strcat(text, "#");
  strcat(text, pid);
  strcat(text, "#");
  strcat(text, dir);


  XChangeProperty(XtDisplay(W_App), XtWindow(W_App), _FOF_SPECIFIC_PROPERTY, _EXODE_CLIENT_PROPERTY_TYPE, 8, PropModeReplace, text, 256+4*128+128);


  /* add the current window to the root property type */

  AddWindowToRootProperty();

}

void ChooseClientCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{
  int num_client;
  XmToggleButtonCallbackStruct *cbs;

  cbs = (XmToggleButtonCallbackStruct *) widget_data;

  if(cbs->set==XmSET) {
    num_client=(int) app_data;

    XMapRaised(XtDisplay(W_App), WindowIds[num_client]);
    XWarpPointer(XtDisplay(W_App), None, WindowIds[num_client], 0,0,0,0,0,0);
  }
}


int NewErrorHandler(Display *disp, XErrorEvent *xerrorev)
{
  return 0;
}


void ActivateMenuCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{
  
  Widget w_temp;
  Display *display;
  int i;
  int result;
  Atom ret_atom;
  int ret_format;
  unsigned long ret_nb_elements;
  unsigned long ret_left;
  unsigned char *propval;
  char exode_sign[128];
  char client_sign[128];
  char windowid[128];
  char pid[128];
  char dir[256];
  Window *windowlist;
  unsigned long nbwindows;
  XmString xmbstring;
  Window wwid;
  unsigned long ppid;

  if(NbButtons!=0) {

    for(i=0;i<NbButtons;i++) {
      XtUnmanageChild(Buttons[i]);
    }

      NbButtons=0;
  }
      
  display=XtDisplay(W_App);

  /* query the current window list */

  windowlist = QueryRootProperty(&nbwindows);

  /* change the X error handler */

  OldErrorHandler = XSetErrorHandler(NewErrorHandler);

  /* search the interesting clients */

    for(i=0;i<nbwindows;i++) {

	result=XGetWindowProperty(display, windowlist[i], _FOF_SPECIFIC_PROPERTY, 0, 128+32, False, _EXODE_CLIENT_PROPERTY_TYPE, &ret_atom, &ret_format, &ret_nb_elements, &ret_left, &propval);
	
	/* if the window has the property */

	if(result!=Success) 
	  RemoveWindowFromRootProperty(windowlist[i]);
	else {
	  if(ret_nb_elements!=0) {
	    
	    sscanf(propval,"%[^#]#%[^#]#%[^#]#%[^#]#%[^#]",exode_sign, client_sign, windowid, pid,dir);
	    sscanf(windowid, "%x", &wwid);
	    sscanf(pid,"%lu", &ppid);
	    
	    WindowIds=(Window*) XtRealloc((char *) WindowIds, sizeof(Window)*(NbButtons+1));
	    WindowIds[NbButtons]=wwid;
	    Pids=(unsigned long *) XtRealloc((char *) Pids, sizeof(unsigned long )*(NbButtons+1));
	  Pids[NbButtons]=ppid;

	  xmbstring=XmStringCreateSimple(dir);

	  if(NbButtons>=MaxButtons) {
	    Buttons=(Widget *) XtRealloc((char *) Buttons, sizeof(Widget)*(NbButtons+1));
	    MaxButtons=NbButtons+1;

	    w_temp = XtNameToWidget(W_MainWindow, "*WindowsMenu");

	    Buttons[NbButtons]= XtVaCreateManagedWidget("ClientBtn", xmToggleButtonWidgetClass, w_temp, XmNlabelString, xmbstring, NULL);

            XtAddCallback(Buttons[NbButtons], XmNvalueChangedCallback, ChooseClientCB, (XtPointer) NbButtons);

	  } else {
	    XtManageChild(Buttons[NbButtons]);
	    XtVaSetValues(Buttons[NbButtons], XmNlabelString, xmbstring, NULL);
	  }

	  if(windowlist[i]==XtWindow(W_App))
	    XtVaSetValues(Buttons[NbButtons], XmNset, True, NULL);
	  else
	    XtVaSetValues(Buttons[NbButtons], XmNset, False, NULL);

	  XmStringFree(xmbstring);

	  NbButtons++;
	}
    
	XFree(propval);
	}
    }

    XSetErrorHandler(OldErrorHandler);

    if(windowlist!=NULL) XFree((unsigned long *) windowlist);

}

void TileButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{

  int i;
  int nb_last_col;
  int nb_windows;
  int max_win;
  int nb_row;
  unsigned int rootwidth;
  unsigned int rootheight;
  unsigned int winwidth;
  unsigned int winheight;
  unsigned int xwin;
  unsigned int ywin;
  Display * display;
  Window root;
  unsigned int b_width;
  unsigned int prof;
  int result;

  display = XtDisplay(W_App);

  nb_windows=NbButtons;
  if(nb_windows==0) return;

  /* calculate the number of rows */

  nb_row= (nb_windows + 1) / 2;

  /* calculate the number of columns on the last row (0 or 1) */

  nb_last_col = nb_windows % 2;

  /* windows sizes */

  result=XGetGeometry(display, DefaultRootWindow(display), &root, &xwin, &ywin, &rootwidth, &rootheight, &b_width, &prof);

  if(result==0) return;

  winwidth=( (rootwidth - RWin_LeftOffset - RWin_RightOffset) - ((nb_row - 1)%2)*RWin_XInterOffset) /2;
  winheight=( (rootheight - RWin_TopOffset - RWin_BottomOffset) - (nb_row - 1)*RWin_YInterOffset) / nb_row;

  /* move the windows */

  if(nb_last_col==1)
    max_win=nb_windows-1;
  else max_win=nb_windows;

  xwin=RWin_LeftOffset;
  ywin=RWin_TopOffset;

  for(i=0;i<max_win;i++) {

    XMoveResizeWindow(display, WindowIds[i], xwin, ywin, winwidth, winheight);
    XRaiseWindow(display, WindowIds[i]);

    xwin+= RWin_XInterOffset + winwidth;
    if(i%2==1) {
      xwin=RWin_LeftOffset;
      ywin+=RWin_YInterOffset + winheight;
    }
  }

  if(nb_last_col==1) {
    XMoveResizeWindow(display, WindowIds[i], xwin, ywin, 2*winwidth+RWin_XInterOffset, winheight);
    XRaiseWindow(display, WindowIds[i]);
  }

}


void StackButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{
  int i;
  int nb_windows;
  unsigned int rootwidth;
  unsigned int rootheight;
  unsigned int winwidth;
  unsigned int winheight;
  unsigned int xwin;
  unsigned int ywin;
  Display * display;
  Window root;
  unsigned int b_width;
  unsigned int prof;
  int result;

  display = XtDisplay(W_App);

  nb_windows=NbButtons;
  if(nb_windows==0) return;

  /* windows sizes */

  result=XGetGeometry(display, DefaultRootWindow(display), &root, &xwin, &ywin, &rootwidth, &rootheight, &b_width, &prof);

  if(result==0) return;

  winwidth= rootwidth - RWin_LeftOffset - RWin_RightOffset;
  winheight=( (rootheight - RWin_TopOffset - RWin_BottomOffset) - (nb_windows - 1)*RWin_YInterOffset) / nb_windows;

  /* move the windows */

  xwin=RWin_LeftOffset;
  ywin=RWin_TopOffset;

  for(i=0;i<nb_windows;i++) {

    XMoveResizeWindow(display, WindowIds[i], xwin, ywin, winwidth, winheight);
    XRaiseWindow(display, WindowIds[i]);

    ywin+=RWin_YInterOffset + winheight;

  }

}

void CascadeButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{
  int i;
  int nb_windows;
  unsigned int rootwidth;
  unsigned int rootheight;
  unsigned int winwidth;
  unsigned int winheight;
  unsigned int xwin;
  unsigned int ywin;
  Display * display;
  Window root;
  unsigned int b_width;
  unsigned int prof;
  int result;

  display = XtDisplay(W_App);

  nb_windows=NbButtons;
  if(nb_windows==0) return;

  /* windows sizes */

  result=XGetGeometry(display, DefaultRootWindow(display), &root, &xwin, &ywin, &rootwidth, &rootheight, &b_width, &prof);

  if(result==0) return;

  winwidth= rootwidth - RWin_LeftOffset - RWin_RightOffset - (nb_windows-1)*RWin_CascadeXOffset;
  winheight= rootheight - RWin_TopOffset - RWin_BottomOffset - (nb_windows-1)*RWin_CascadeYOffset;

  /* move the windows */

  xwin=RWin_LeftOffset+(nb_windows-1)*RWin_CascadeXOffset;
  ywin=RWin_TopOffset+(nb_windows-1)*RWin_CascadeYOffset;

  for(i=nb_windows-1;i>=0;i--) {
    XMoveResizeWindow(display, WindowIds[nb_windows-1-i], xwin, ywin, winwidth, winheight);
    XRaiseWindow(display, WindowIds[nb_windows-1-i]);
    xwin-=RWin_CascadeXOffset;
    ywin-=RWin_CascadeYOffset;
  }

}

void QuitOthersButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{

  int i;
  pid_t current_pid;

  ActivateMenuCB(NULL, NULL, NULL);

  current_pid=getpid();

  for(i=0;i<NbButtons;i++) {
    if(Pids[i]!=current_pid) 
      kill(Pids[i], SIGKILL);
    if(WindowIds[i]!=XtWindow(W_App))
      RemoveWindowFromRootProperty(WindowIds[i]);
  }

}


void QuitAllButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{

  int i;
  pid_t current_pid;

  ActivateMenuCB(NULL, NULL, NULL);

  current_pid=getpid();

  for(i=0;i<NbButtons;i++) {
    if(Pids[i]!=current_pid) 
      kill(Pids[i], SIGKILL);
  }

  RemoveRootProperty();

  XFlush(XtDisplay(W_App));

  exit(EXIT_SUCCESS);
}

void InstallWindowMenuCB(void)
{
  Widget w;

  /* Tile Button */

  w = XtNameToWidget(W_App, "*TileButton");
  XtAddCallback(w, XmNactivateCallback, TileButtonCB, NULL);

  /* Stack Button */

  w = XtNameToWidget(W_App, "*StackButton");
  XtAddCallback(w, XmNactivateCallback, StackButtonCB, NULL);

  /* Cascade Button */

  w = XtNameToWidget(W_App, "*CascadeButton");
  XtAddCallback(w, XmNactivateCallback, CascadeButtonCB, NULL);

  /* Quit Others Button */

  w = XtNameToWidget(W_App, "*QuitOthersButton");
  XtAddCallback(w, XmNactivateCallback, QuitOthersButtonCB, NULL);

  /* Quit All Button */

  w = XtNameToWidget(W_App, "*QuitAllButton");
  XtAddCallback(w, XmNactivateCallback, QuitAllButtonCB, NULL);

  /* Window menu */

  w = XtNameToWidget(W_App, "*WindowsMenu");
  XtAddCallback(w, XmNmapCallback, ActivateMenuCB, NULL);
}

void DeleteWindowCallback(Widget w, XtPointer clientdata, XtPointer calldata)
{
  RemoveWindowFromRootProperty(XtWindow(W_App));
  XFlush(XtDisplay(W_App));
  exit(0);
}

void InitModuleWindows(void)
{
  Atom WM_DELETE_WINDOW;


  WM_DELETE_WINDOW = XInternAtom( XtDisplay(W_App), "WM_DELETE_WINDOW", False);

  XmAddWMProtocolCallback( W_App, WM_DELETE_WINDOW, DeleteWindowCallback, NULL);

  InstallWindowMenuCB();

  GetVarInt("WinLeftOffset", &RWin_LeftOffset);
  GetVarInt("WinTopOffset", &RWin_TopOffset);
  GetVarInt("WinRightOffset", &RWin_RightOffset);
  GetVarInt("WinBottomOffset", &RWin_BottomOffset);
  GetVarInt("WinXInterOffset", &RWin_XInterOffset);
  GetVarInt("WinYInterOffset", &RWin_YInterOffset);
  GetVarInt("WinXCascadeOffset", &RWin_CascadeXOffset);
  GetVarInt("WinYCascadeOffset", &RWin_CascadeYOffset);
}











